/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/common.h: Base header used by all files.
 */

#ifndef COMMON_H
#define COMMON_H

/*************************************************************************/
/************************* Configuration options *************************/
/*************************************************************************/

/*
 * DEBUG
 *
 * When defined, the program is compiled in debugging mode, with the
 * following changes:
 *
 *    - Failures of preconditions (the PRECOND() macro) will cause the
 *      program to abort.
 *
 *    - Debug messages will be written to standard error.
 */

#define DEBUG

/*-----------------------------------------------------------------------*/

/*
 * MAX_REGIONS
 *
 * Specifies the maximum number of regions into which a wave file may be
 * divided.  Roughly, this is the maximum number of operations which may
 * be performed on a wave file without saving it to disk.
 */

#define MAX_REGIONS  1000

/*************************************************************************/
/************************* Common system headers *************************/
/*************************************************************************/

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/*************************************************************************/
/*************************** Common data types ***************************/
/*************************************************************************/

/*
 * CanvasWindow:  Type of a wave file canvas window data structure.
 */
typedef struct CanvasWindow_ CanvasWindow;

/*
 * WaveFile:  Type of a wave file data structure.
 */
typedef struct WaveFile_ WaveFile;

/*************************************************************************/
/***************************** Useful macros *****************************/
/*************************************************************************/

/**** Integer macros ****/

/* Minimum, maximum, and absolute value. */
#undef min
#undef max
#undef abs
#define min(a,b)  ((a) < (b) ? (a) : (b))
#define max(a,b)  ((a) > (b) ? (a) : (b))
#define abs(n)    ((n) > 0 ? (n) : -(n))

/* Bounding macros.  Functionally equivalent to min() and max(), but these
 * names help make the code more readable when the intent is to bound a
 * value within specified limits. */
#define lbound(n,lower)       max((n), (lower))
#define ubound(n,upper)       min((n), (upper))
#define bound(n,lower,upper)  ubound(lbound((n), (lower)), (upper))

/* Align x up or down to a multiple of "align".  Note that GCC 4.2+ is
 * capable of optimizing "a/b*b" (where b is a power of 2) to "a&-b", so
 * this general form doesn't hurt speed in optimizable cases. */
#define align_up(x,align)    (((x) + ((align)-1)) / (align) * (align))
#define align_down(x,align)  ((x) / (align) * (align))

/**** Miscellaneous macros ****/

/* Return the length (in elements) of the given array. */
#define lenof(array)  ((int)(sizeof(array) / sizeof(*(array))))

/* Declare a function as taking a printf()-style format string and
 * arguments. */
#ifdef __GNUC__
# define FORMAT(fmt,firstarg) __attribute__((format(printf,fmt,firstarg)))
#else
# define FORMAT(fmt,firstarg) /*nothing*/
#endif

/*************************************************************************/
/*********************** Debugging/testing macros ************************/
/*************************************************************************/

/*
 * DMSG(format,args...)
 *
 * Displays a formatted debug message, prepending the source file, line
 * number, and function name, and appending a newline.  The format string
 * must be a string literal.
 *
 * Does nothing if DEBUG is not defined.
 */

#ifdef DEBUG
# define DMSG(format,...)  fprintf(stderr, "%s:%d(%s): " format "\n", \
                                   __FILE__, __LINE__, __FUNCTION__ \
                                   , ## __VA_ARGS__)
#else
# define DMSG(format,...)  /*nothing*/
#endif

/*************************************************************************/

/*
 * PRECOND(condition,fail_action)
 *
 * If DEBUG is defined, tests whether the given condition evaluates to
 * true, and aborts the program with an error message if not.
 *
 * If DEBUG is not defined, executes fail_action if the condition
 * evaluates to false,
 */

#ifdef DEBUG

# define PRECOND(condition,fail_action)  do {                           \
    if (!(condition)) {                                                 \
        fprintf(stderr,                                                 \
                "\n*** ALERT *** PRECONDITION FAILED\n%s:%d(%s): %s\n\n", \
                __FILE__, __LINE__, __FUNCTION__, #condition);          \
        /* fail_action will not be executed, but compile it anyway so that \
         * syntax errors and the like are still caught in debug mode. */ \
        if (0) {fail_action;}                                           \
        abort();                                                        \
    }                                                                   \
} while (0)

#else  // !DEBUG

# define PRECOND(condition,fail_action)  do {   \
    if (!(condition)) {                         \
        fail_action;                            \
    }                                           \
} while (0)

#endif

/*************************************************************************/
/*************************************************************************/

#endif  // COMMON_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
