/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/manager.c: Interface for managing all loaded/created wave files.
 */

#include "common.h"
#include "manager.h"
#include "ui.h"

#include <glib.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* List of all loaded wave files. */
static GSList *wave_list = NULL;

/* Iterator for first()/next(). */
static GSList *wave_list_iterator = NULL;

/* Pointer to currently active wave file. */
static WaveFile *active_wave = NULL;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

/**
 * manager_add_wave:  Add a wave file to the list of loaded files.
 *
 * [Parameters]
 *     wave: Wave file to add
 * [Return value]
 *     None
 */
void manager_add_wave(WaveFile *wave)
{
    wave_list = g_slist_append(wave_list, wave);
}

/*----------------------------------*/

/**
 * manager_delete_wave:  Delete a wave file to the list of loaded files.
 *
 * [Parameters]
 *     wave: Wave file to delete
 * [Return value]
 *     None
 */
void manager_delete_wave(WaveFile *wave)
{
    GSList *node = g_slist_find(wave_list, wave);
    if (node) {
        if (wave == active_wave) {
            /* We're deleting the active wave file.  Move to the next one
             * in the list, or the first one if this was the last. */
            GSList *next_node = g_slist_next(node);
            if (next_node) {
                manager_set_active_wave(next_node->data);
            } else if (wave_list != node) {
                manager_set_active_wave(wave_list->data);
            } else {
                /* This is the only wave file left. */
                manager_set_active_wave(NULL);
            }
        }
        wave_list = g_slist_delete_link(wave_list, node);
    }
}

/*************************************************************************/

/**
 * manager_first_wave:  Return the first wave file in the list of loaded
 * files.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     First loaded wave file
 */
WaveFile *manager_first_wave(void)
{
    wave_list_iterator = wave_list;
    return manager_next_wave();
}

/*----------------------------------*/

/**
 * manager_next_wave:  Return the next wave file in the list of loaded
 * files after the last call to manager_first_wave() or manager_next_wave().
 * If a wave file is added to the list, or any wave file but the last
 * returned from manager_first_wave() or manager_next_wave() is deleted
 * from the list, the behavior of this function is undefined.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Next loaded wave file
 */
WaveFile *manager_next_wave(void)
{
    WaveFile *next = wave_list_iterator ? wave_list_iterator->data : NULL;
    wave_list_iterator = g_slist_next(wave_list_iterator);
    return next;
}

/*************************************************************************/

/**
 * manager_set_active_wave:  Set the currently active wave file.
 *
 * [Parameters]
 *     wave: Wave file to set as active, or NULL to set no wave file as active
 * [Return value]
 *     None
 */
void manager_set_active_wave(WaveFile *wave)
{
    active_wave = wave;
    ui_update();
}

/*----------------------------------*/

/**
 * manager_get_active_wave:  Return the currently active wave file.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Currently active wave file, or NULL if none
 */
WaveFile *manager_get_active_wave(void)
{
    return active_wave;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
