/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/wave.c: Interface for using and editing individual wave files.
 */

#include "common.h"
#include "canvas.h"
#include "wave.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Wave data region structure.  Unsaved changes to wave data are tracked by
 * an array of these structures which indicates the current state of the
 * wave file. */

typedef struct WaveRegion_ WaveRegion;

struct WaveRegion_ {

    /* Sample index of the first sample in this region.  This is only a
     * convenience field; its value is equal to the sum of the "length"
     * field in all preceding regions. */
    uint64_t start;

    /* Length of this region, in samples. */
    uint64_t length;

    /* Data source for this region:
     *    - If not NULL, this points to an allocated buffer containing the
     *      sample data for this region.  The data is stored as floating
     *      point regardless of the file's sample format.
     *    - If NULL, the data for this region is stored in the backing file
     *      at the sample offset given in "file_offset".
     * This field must be non-NULL if there is no backing file, i.e. if
     * wave->file is NULL. */
    float *data;

    /* File offset for file-backed regions, in samples from the start of
     * the sample data in the file. */
    uint64_t file_offset;

};  // struct WaveRegion_

/*----------------------------------*/

/* Wave file data structure.  The structure is only defined here so as to
 * render it opaque to callers. */

struct WaveFile_ {

    /* Pathname for this wave file (allocated), or NULL if none, e.g. if
     * this wave file was created as a new file or recorded. */
    char *path;

    /* Disk file backing this wave file's data, or NULL if none. */
    FILE *file;

    /* Byte offset of data in backing file.  This field is ignored if
     * file is NULL. */
    uint64_t data_offset;

    /* Is the disk file writable?  This field is ignored if file is NULL. */
    uint8_t writable;

    /* Has this wave file been modified since last loaded or saved? */
    uint8_t modified;

    /*----------------*/

    /* Wave data format. */
    WaveFormat format;
    unsigned int bits;
    unsigned int channels;
    unsigned int srate;

    /* Wave length in samples. */
    uint64_t samples;

    /* Region list and count. */
    WaveRegion regions[MAX_REGIONS];
    unsigned int num_regions;

    /*----------------*/

    /* Cached return data for wave_get_full_canvas(). */
    unsigned int full_canvas_cache_width;
    uint16_t *full_canvas_cache;

    /* Canvas window for this wave file. */
    CanvasWindow *canvas;

    /* Current and saved cursor positions, in samples. */
    uint64_t position;
    uint64_t saved_position;

    /* Current and saved selection length, in samples. */
    uint64_t select_len;
    uint64_t saved_select_len;

    /* Zoom factor for canvas window.  The canvas window will display this
     * many samples per pixel. */
    uint64_t zoom;

};  // struct WaveFile

/*************************************************************************/

/* Local function declarations. */

static unsigned int find_region(const WaveFile *wave, uint64_t position);
static int split_region(WaveFile *wave, uint64_t position);
static void update_region_starts(WaveFile *wave);

static void find_high_low(const WaveFile *wave, uint64_t start,
                          uint64_t length, uint16_t *data_ret);
static void find_high_low_region(const WaveFile *wave, unsigned int region,
                                 uint64_t offset, uint64_t length,
                                 uint16_t *data_ret);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

/**
 * wave_create:  Create a new, empty wave file.
 *
 * [Parameters]
 *       format: Format of the wave data (FMT_*)
 *     channels: Number of channels in the wave file (1-6)
 *        srate: Sampling rate of the wave data (at least 1)
 * [Return value]
 *     Newly-created wave file, or NULL on error
 */
WaveFile *wave_create(WaveFormat format, unsigned int channels,
                      unsigned int srate)
{
    PRECOND(channels >= 1 && channels <= 6, return NULL);
    PRECOND(srate >= 1, return NULL);

    WaveFile *wave = malloc(sizeof(*wave));
    if (!wave) {
        DMSG("Out of memory");
        return NULL;
    }

    wave->path              = NULL;
    wave->file              = NULL;
    wave->data_offset       = 0;  // Ignored, but set it anyway
    wave->writable          = 1;  // Ignored, but set it anyway
    wave->modified          = 0;

    wave->format            = format;
    wave->bits              = 8*format;
    wave->channels          = channels;
    wave->srate             = srate;
    wave->samples           = 0;
    wave->num_regions       = 0;

    wave->full_canvas_cache_width = 0;
    wave->full_canvas_cache = NULL;

    wave->position          = 0;
    wave->saved_position    = 0;
    wave->select_len        = 0;
    wave->saved_select_len  = 0;
    wave->zoom              = 1;

    /* Create the canvas window for the wave file.  This has to be done
     * last because the canvas will call back to us to read wave data. */
    wave->canvas = canvas_create(wave);
    if (!wave->canvas) {
        DMSG("Failed to create canvas");
        free(wave);
        return NULL;
    }

    return wave;
}

/*-----------------------------------------------------------------------*/

/**
 * wave_load:  Load a .WAV file from disk, returning the associated wave
 * file structure.
 *
 * [Parameters]
 *     path: Pathname of file to load
 * [Return value]
 *     Loaded wave file, or NULL on error
 */
WaveFile *wave_load(const char *path)
{
    PRECOND(path != NULL, return NULL);

    WaveFile *wave = malloc(sizeof(*wave));
    if (!wave) {
        DMSG("Out of memory");
        return NULL;
    }
    wave->path = strdup(path);
    if (!wave->path) {
        DMSG("Out of memory");
        free(wave);
        return NULL;
    }

    /* Initialize other fields early so we can call wave_destroy() on error. */
    wave->modified          = 0;
    wave->num_regions       = 0;
    wave->full_canvas_cache_width = 0;
    wave->full_canvas_cache = NULL;
    wave->canvas            = NULL;
    wave->position          = 0;
    wave->saved_position    = 0;
    wave->select_len        = 0;
    wave->saved_select_len  = 0;
    wave->zoom              = 1;

    /* First try to open the file read-write, so we know we can write data
     * to it at wave_save() time.  If that doesn't work, try opening in
     * read-only mode instead. */
    wave->file = fopen(path, "r+b");
    if (wave->file) {
        wave->writable = 1;
    } else {
        wave->file = fopen(path, "rb");
        if (wave->file) {
            wave->writable = 0;
        } else {
            DMSG("Failed to open %s: %s", path, strerror(errno));
            wave_destroy(wave);
            return NULL;
        }
    }

    /* Check that the file is seekable, just to make sure we're not trying
     * to open a named pipe or some such. */
    if (fseek(wave->file, SEEK_END, 0) != 0
     || fseek(wave->file, SEEK_SET, 0) != 0
    ) {
        DMSG("%s is not seekable (fseek(): %s)", path, strerror(errno));
        wave_destroy(wave);
        return NULL;
    }

    /* Check that the file is in RIFF WAVE format, and parse the file
     * header to obtain the wave data format. */
    uint8_t buf[12];
    if (fread(buf, 1, 12, wave->file) != 12) {
        DMSG("Failed to scan %s (premature EOF on RIFF header)", path);
        wave_destroy(wave);
        return NULL;
    }
    int64_t loopstart = -1, looplen = -1;
    if (memcmp(buf, "RIFF", 4) == 0 && memcmp(buf+8, "WAVE", 4) == 0) {
        wave->bits = 0;  // Use this to ensure we've found the "fmt " chunk.
        uint32_t chunksize;
        for (;;) {
            if (fread(buf, 1, 8, wave->file) != 8) {
                DMSG("Failed to scan %s (premature EOF on chunk header)", path);
                wave_destroy(wave);
                return NULL;
            }
            chunksize = buf[4] | buf[5]<<8 | buf[6]<<16 | buf[7]<<24;
            if (memcmp(buf, "data", 4) == 0) {
                break;
            } else if (memcmp(buf, "fmt ", 4) == 0) {
                if (chunksize < 16) {
                    DMSG("Failed to scan %s (fmt chunk too small)", path);
                    wave_destroy(wave);
                    return NULL;
                }
                uint8_t fmt[40];
                const int readsize = ubound(chunksize, sizeof(fmt));
                if (fread(fmt, 1, readsize, wave->file) != readsize) {
                    DMSG("Failed to scan %s (premature EOF on fmt chunk)",
                         path);
                    wave_destroy(wave);
                    return NULL;
                }
                if (chunksize > readsize) {
                    if (fseek(wave->file, chunksize - readsize,
                              SEEK_CUR) != 0) {
                        DMSG("Failed to scan %s (fseek(%u,SEEK_CUR): %s)",
                             path, chunksize - readsize, strerror(errno));
                        wave_destroy(wave);
                        return NULL;
                    }
                }
                uint16_t codec = fmt[0] | fmt[1]<<8;
                const uint16_t channels = fmt[2] | fmt[3]<<8;
                const uint32_t srate =
                    fmt[4] | fmt[5]<<8 | fmt[6]<<16 | fmt[7]<<24;
                const uint16_t bits = fmt[14] | fmt[15]<<8;
                if (codec == 0xFFFE) {
                    if (chunksize < 40) {
                        DMSG("Failed to scan %s (fmt chunk too small for"
                             " extended data)", path);
                        wave_destroy(wave);
                        return NULL;
                    }
                    codec = fmt[24] | fmt[25]<<8;
                }
                if (codec != 0x0001 && codec != 0x0003) {
                    DMSG("%s is not a linear PCM file", path);
                    wave_destroy(wave);
                    return NULL;
                }
                if ((channels < 1 || channels > 6)
                 || (srate == 0)
                 || (codec == 0x0001 && bits != 8 && bits != 16 && bits != 24)
                 || (codec == 0x0003 && bits != 32)
                ) {
                    DMSG("Unsupported PCM format in %s: %u/%u/%u", path,
                         srate, channels, bits);
                    wave_destroy(wave);
                    return NULL;
                }
                if (codec == 0x0001) {
                    if (bits == 8) {
                        wave->format = FMT_U8;
                    } else if (bits == 16) {
                        wave->format = FMT_S16;
                    } else {
                        assert(bits == 24);
                        wave->format = FMT_S24;
                    }
                } else {
                    assert(codec == 0x0003);
                    assert(bits == 32);
                    wave->format = FMT_F32;
                }
                wave->bits = bits;
                wave->channels = channels;
                wave->srate = srate;
            } else if (memcmp(buf, "smpl", 4) == 0 && chunksize >= 60) {
                uint8_t smpl[60];
                const int readsize = sizeof(smpl);
                if (fread(smpl, 1, readsize, wave->file) != readsize) {
                    DMSG("Failed to scan %s (premature EOF on smpl chunk)",
                         path);
                    wave_destroy(wave);
                    return NULL;
                }
                if (chunksize > readsize) {
                    if (fseek(wave->file, chunksize - readsize,
                              SEEK_CUR) != 0) {
                        DMSG("Failed to scan %s (fseek(%u,SEEK_CUR): %s)",
                             path, chunksize - readsize, strerror(errno));
                        wave_destroy(wave);
                        return NULL;
                    }
                }
                const int num_loops =
                    smpl[28] | smpl[29]<<8 | smpl[30]<<16 | smpl[31]<<24;
                const int start =
                    smpl[44] | smpl[45]<<8 | smpl[46]<<16 | smpl[47]<<24;
                const int end =
                    smpl[48] | smpl[49]<<8 | smpl[50]<<16 | smpl[51]<<24;
                if (num_loops == 1 && end >= start) {
                    loopstart = start;
                    looplen = end - start + 1;
                }
            } else {  // Not "fmt ", "smpl", or "data", so skip.
                if (fseek(wave->file, chunksize, SEEK_CUR) != 0) {
                    DMSG("Failed to scan %s (fseek(%u,SEEK_CUR): %s)",
                         path, chunksize, strerror(errno));
                    wave_destroy(wave);
                    return NULL;
                }
            }
        }
        if (wave->bits == 0) {
            DMSG("Failed to scan %s (missing 'fmt ' chunk)", path);
            wave_destroy(wave);
            return NULL;
        }
        /* If we get here, we must be at a "data" chunk, so determine the
         * wave data length and save the file offset. */
        wave->samples = chunksize / (wave->bits/8 * wave->channels);
        wave->data_offset = ftell(wave->file);
    } else {
        DMSG("%s is not a RIFF WAVE", path);
        wave_destroy(wave);
        return NULL;
    }

    /* Make a single region covering the entire file.  (But if it's an
     * empty file, don't make any regions at all.) */
    if (wave->samples > 0) {
        wave->regions[0].start       = 0;
        wave->regions[0].length      = wave->samples;
        wave->regions[0].data        = NULL;
        wave->regions[0].file_offset = 0;
        wave->num_regions = 1;
        if (loopstart >= 0) {
            if (loopstart > wave->samples - 1) {
                looplen -= loopstart - (wave->samples - 1);
                loopstart = wave->samples - 1;
            }
            if (looplen > wave->samples - loopstart) {
                looplen = wave->samples - loopstart;
            }
            /* Set the selection to the loop so it can be played. */
            wave->position = loopstart;
            wave->select_len = looplen;
        }
    } else {
        wave->num_regions = 0;
    }

    /* Create the canvas window for the wave file. */
    wave->canvas = canvas_create(wave);
    if (!wave->canvas) {
        DMSG("Failed to create canvas");
        wave_destroy(wave);
        return NULL;
    }

    return wave;
}

/*-----------------------------------------------------------------------*/

/**
 * wave_save:  Save a wave file to disk.  The newly-created file will be
 * used for subsequent access.
 *
 * [Parameters]
 *     wave: Wave file to save
 *     path: New pathname for file
 * [Return value]
 *     Nonzero on success, zero on error
 */
int wave_save(WaveFile *wave, const char *path)
{
    PRECOND(wave != NULL, return 0);
    PRECOND(path != NULL, return 0);

    /* Make a copy of the pathname, since we'll need it to store in
     * wave->path when we're done. */
    char *path_copy = strdup(path);
    if (!path_copy) {
        DMSG("Out of memory");
        goto error_return;
    }

    /* Create a temporary file that can later be renamed over the original. */
    const unsigned int tempname_bufsize = strlen(path) + 7;
    char *tempname_buf = malloc(tempname_bufsize);
    if (!tempname_buf) {
        DMSG("Out of memory");
        goto error_free_path_copy;
    }
    snprintf(tempname_buf, tempname_bufsize, "%sXXXXXX", path);
    int newfd = mkstemp(tempname_buf);
    if (newfd < 0) {
        DMSG("mkstemp(%s): %s", tempname_buf, strerror(errno));
        goto error_free_tempname;
    }

    /* Create a stdio handle for the new file. */
    FILE *newfile = fdopen(newfd, "r+b");
    if (!newfile) {
        DMSG("fdopen(%d): %s", newfd, strerror(errno));
        close(newfd);
        goto error_remove_tempname;
    }

    /* Write the file header.  If the length overflows 32 bits, we just
     * truncate. */
    uint8_t header[68] = "RIFF....WAVEfmt \x10\0\0\0\1\0";
    const uint64_t samplesize = (wave->bits/8) * wave->channels;
    const uint64_t datarate = samplesize * wave->srate;
    const uint64_t datasize = samplesize * wave->samples;
    uint64_t filesize = datasize + (44-8);
    header[22] = wave->channels >>  0 & 0xFF;
    header[23] = wave->channels >>  8 & 0xFF;
    header[24] = wave->srate    >>  0 & 0xFF;
    header[25] = wave->srate    >>  8 & 0xFF;
    header[26] = wave->srate    >> 16 & 0xFF;
    header[27] = wave->srate    >> 24 & 0xFF;
    header[28] = datarate       >>  0 & 0xFF;
    header[29] = datarate       >>  8 & 0xFF;
    header[30] = datarate       >> 16 & 0xFF;
    header[31] = datarate       >> 24 & 0xFF;
    header[32] = samplesize     >>  0 & 0xFF;
    header[33] = samplesize     >>  8 & 0xFF;
    header[34] = wave->bits     >>  0 & 0xFF;
    header[35] = wave->bits     >>  8 & 0xFF;
    int dataofs = 36;
    if (wave->channels > 2) {
        filesize += 24;
        dataofs += 24;
        header[16] += 24;
        header[20] = 0xFE;
        header[21] = 0xFF;
        memcpy(&header[36], "\x16\0\x10\0\0\0\0\0\1\0\0\0\0\0\x10\0"
                            "\x80\0\0\xAA\0\x38\x9B\x71", 24);
        header[40] = (1 << wave->channels) - 1;
    }
    header[ 4] = filesize >>  0 & 0xFF;
    header[ 5] = filesize >>  8 & 0xFF;
    header[ 6] = filesize >> 16 & 0xFF;
    header[ 7] = filesize >> 24 & 0xFF;
    memcpy(&header[dataofs], "data", 4);
    header[dataofs+4] = datasize >>  0 & 0xFF;
    header[dataofs+5] = datasize >>  8 & 0xFF;
    header[dataofs+6] = datasize >> 16 & 0xFF;
    header[dataofs+7] = datasize >> 24 & 0xFF;
    if (fwrite(header, dataofs+8, 1, newfile) != 1) {
        DMSG("fwrite(header): %s", strerror(errno));
        goto error_close_newfile;
    }

    /* Copy the current wave data to the new file. */

    uint64_t position;
    uint32_t count;  // Declare here so we can use it in the for() increment.

    for (position = 0; position < wave->samples; position += count) {

        /* Get the next buffer's worth of samples. */
        uint8_t buf[65536];
        count = ubound(wave->samples - position, sizeof(buf) / samplesize);
        float *data = wave_get_data(wave, position, count);
        if (!data) {
            DMSG("Failed to get data at %llu", (unsigned long long)position);
            goto error_close_newfile;
        }

        /* Convert to the desired output format and free the input data. */
        const float *src = data;
        uint8_t *dest = buf;
        switch (wave->format) {
          case FMT_U8:
            for (uint32_t i = 0; i < count; i++) {
                unsigned int ch;
                for (ch = 0; ch < wave->channels; ch++, src++, dest++) {
                    *dest = bound((int)roundf(*src * 128), -128, 127) ^ 0x80;
                }
            }
            break;
          case FMT_S16:
            for (uint32_t i = 0; i < count; i++) {
                unsigned int ch;
                for (ch = 0; ch < wave->channels; ch++, src++, dest += 2) {
                    int value = bound((int)roundf(*src * 32768), -32768, 32767);
                    dest[0] = value>>0 & 0xFF;
                    dest[1] = value>>8 & 0xFF;
                }
            }
            break;
          case FMT_S24:
            for (uint32_t i = 0; i < count; i++) {
                unsigned int ch;
                for (ch = 0; ch < wave->channels; ch++, src++, dest += 3) {
                    int value = bound((int)roundf(*src * 8388608),
                                      -8388608, 8388607);
                    dest[0] = value>>0 & 0xFF;
                    dest[1] = value>>8 & 0xFF;
                    dest[2] = value>>16 & 0xFF;
                }
            }
            break;
          case FMT_F32:
            for (uint32_t i = 0; i < count; i++) {
                unsigned int ch;
                for (ch = 0; ch < wave->channels; ch++, src++, dest += 4) {
                    union {float f; uint32_t i;} value = {.f = *src};
                    dest[0] = value.i>>0 & 0xFF;
                    dest[1] = value.i>>8 & 0xFF;
                    dest[2] = value.i>>16 & 0xFF;
                    dest[3] = value.i>>24 & 0xFF;
                }
            }
            break;
        }
        free(data);

        /* Write the converted data to the new file. */
        if (fwrite(buf, samplesize, count, newfile) != count) {
            DMSG("Failed to write data at %llu", (unsigned long long)position);
            goto error_close_newfile;
        }

    }  // for (position = 0; position < wave->samples; position += count)

    /* Flush out any stdio-buffered data. */
    fflush(newfile);

    /* Rename the new file to its final name. */
    if (rename(tempname_buf, path) != 0) {
        DMSG("rename(\"%s\", \"%s\"): %s", tempname_buf, path,
             strerror(errno));
        goto error_close_newfile;
    }

    /* Save successful!  Close the old file handle (if any) and update the
     * wave file structure. */
    if (wave->file) {
        fclose(wave->file);
    }
    free(wave->path);
    wave->path        = path_copy;
    wave->file        = newfile;
    wave->data_offset = dataofs+8;
    wave->writable    = 1;
    wave->modified    = 0;
    /* Don't forget to free all the memory-backed regions. */
    unsigned int region;
    for (region = 0; region < wave->num_regions; region++) {
        free(wave->regions[region].data);
        wave->regions[region].data = NULL;
    }
    if (wave->samples > 0) {
        wave->regions[0].start       = 0;
        wave->regions[0].length      = wave->samples;
        wave->regions[0].data        = NULL;
        wave->regions[0].file_offset = 0;
        wave->num_regions = 1;
    } else {
        wave->num_regions = 0;
    }

    /* Also refresh the canvas window to show any change in filename. */
    canvas_refresh(wave->canvas, 0);

    return 1;

  error_close_newfile:
    fclose(newfile);
  error_remove_tempname:
    remove(tempname_buf);
  error_free_tempname:
    free(tempname_buf);
  error_free_path_copy:
    free(path_copy);
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

/**
 * wave_destroy:  Destroy a wave file and all associated resources.  Does
 * nothing if wave == NULL.
 *
 * [Parameters]
 *     wave: Wave file to destroy
 * [Return value]
 *     None
 */
void wave_destroy(WaveFile *wave)
{
    if (wave) {
        free(wave->path);
        if (wave->file) {
            fclose(wave->file);
        }
        unsigned int region;
        for (region = 0; region < wave->num_regions; region++) {
            free(wave->regions[region].data);
        }
        free(wave->full_canvas_cache);
        canvas_destroy(wave->canvas);
        free(wave);
    }
}

/*************************************************************************/

/**
 * wave_path:  Return the pathname associated with the given wave file, or
 * NULL if the wave file does not have an associated pathname.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Associated pathname, or NULL if none
 */
const char *wave_path(const WaveFile *wave)
{
    PRECOND(wave != NULL, return NULL);
    return wave->path;
}

/*----------------------------------*/

/**
 * wave_writable:  Return whether the disk file backing the wave data for
 * the given wave file is writable.  Always returns true for wave files
 * which are not backed by a disk file, such as newly-created empty files.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     True (nonzero) if the disk file is writable or there is no
 *     associated disk file, else false (zero)
 */
int wave_writable(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 1);
    return !wave->file || wave->writable;
}

/*----------------------------------*/

/**
 * wave_modified:  Returns whether the given wave file has been modified
 * since it was last loaded or saved.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     True (nonzero) if the wave file has been modified since last loaded
 *     or saved, else false (zero)
 */
int wave_modified(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 0);
    return wave->modified;
}

/*----------------------------------*/

/**
 * wave_format, wave_bits, wave_channels, wave_srate, wave_samples:  Return
 * the format, bit resolution, number of channels, sampling rate, or number
 * of samples, respectively, for the given wave file.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Relevant wave file parameter
 */
WaveFormat wave_format(const WaveFile *wave)
{
    PRECOND(wave != NULL, return FMT_S16);
    return wave->format;
}

unsigned int wave_bits(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 16);
    return wave->bits;
}

unsigned int wave_channels(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 2);
    return wave->channels;
}

unsigned int wave_srate(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 48000);
    return wave->srate;
}

uint64_t wave_samples(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 0);
    return wave->samples;
}

/*----------------------------------*/

/**
 * wave_position, wave_select_len, wave_zoom:  Return the current cursor
 * position, selection length, or zoom factor, respectively for the given
 * wave file.  All return values are in units of samples.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Relevant wave file parameter
 */
uint64_t wave_position(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 0);
    return wave->position;
}

uint64_t wave_select_len(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 0);
    return wave->select_len;
}

uint64_t wave_zoom(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 1);
    return wave->zoom;
}

/*----------------------------------*/

/**
 * wave_canvas_width:  Return the width of the wave file's canvas window,
 * in pixels.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Width of canvas window
 */
unsigned int wave_canvas_width(const WaveFile *wave)
{
    PRECOND(wave != NULL, return 1);
    return canvas_width(wave->canvas);
}

/*************************************************************************/

/**
 * wave_set_position, wave_set_select_len, wave_set_zoom:  Set the current
 * cursor position, selection length, or zoom factor, respectively for the
 * given wave file.  All values are in units of samples.
 *
 * [Parameters]
 *            wave: Wave file from which to return data
 *           value: New value for relevant wave file parameter
 *     move_canvas: Nonzero to cause the wave file's canvas window to be
 *                     moved based on the new value, else zero (only for
 *                     set_position() and set_select_len())
 * [Return value]
 *     None
 */
void wave_set_position(WaveFile *wave, uint64_t value, int move_canvas)
{
    PRECOND(wave != NULL, return);
    PRECOND(value <= wave->samples - wave->select_len,
            value = wave->samples - wave->select_len);
    wave->position = value;
    canvas_refresh(wave->canvas, move_canvas);
}

void wave_set_select_len(WaveFile *wave, uint64_t value, int move_canvas)
{
    PRECOND(wave != NULL, return);
    PRECOND(value <= wave->samples - wave->position,
            value = wave->samples - wave->position);
    wave->select_len = value;
    canvas_refresh(wave->canvas, move_canvas);
}

void wave_set_zoom(WaveFile *wave, uint64_t value)
{
    PRECOND(wave != NULL, return);
    PRECOND(value >= 1 && value <= wave->samples, value = 1);
    wave->zoom = value;
    canvas_refresh(wave->canvas, 0);
}

/*----------------------------------*/

/**
 * wave_set_selection:  Set the current cursor position and selection
 * length for the given wave file.  This is a convenience function that
 * wraps set_position() and set_select_len(), and relieves callers from
 * having to worry about the proper order in which to set the values so
 * the selection does not run over the end of the file.
 *
 * [Parameters]
 *            wave: Wave file from which to return data
 *        position: New cursor position, in samples
 *      select_len: New selection length, in samples
 *     move_canvas: Nonzero to cause the wave file's canvas window to be
 *                     moved based on the new values, else zero
 * [Return value]
 *     None
 */
void wave_set_selection(WaveFile *wave, uint64_t position,
                        uint64_t select_len, int move_canvas)
{
    PRECOND(wave != NULL, return);
    PRECOND(position <= wave->samples, position = wave->samples);
    PRECOND(select_len <= wave->samples - position,
            select_len = wave->samples - position);
    wave->position = position;
    wave->select_len = select_len;
    canvas_refresh(wave->canvas, move_canvas);
}

/*----------------------------------*/

/**
 * wave_swap_saved_position:  Swap the current and saved positions and
 * selection lengths for the given wave file.
 *
 * [Parameters]
 *     wave: Wave file for which to swap positions
 * [Return value]
 *     None
 */
void wave_swap_saved_position(WaveFile *wave)
{
    PRECOND(wave != NULL, return);
    uint64_t temp;
    temp = wave->position;
        wave->position = wave->saved_position;
        wave->saved_position = temp;
    temp = wave->select_len;
        wave->select_len = wave->saved_select_len;
        wave->saved_select_len = temp;
    canvas_refresh(wave->canvas, 1);
}

/*************************************************************************/

/**
 * wave_get_data:  Return sample data from the specified portion of the
 * wave file.
 *
 * [Parameters]
 *       wave: Wave file from which to return data
 *      start: Index of first sample to return
 *     length: Number of samples to return
 * [Return value]
 *     Sample data as floating-point values in sample-major, channel-minor
 *     order, or NULL on error
 */
float *wave_get_data(const WaveFile *wave, uint64_t start, uint64_t length)
{
    PRECOND(wave != NULL, return NULL);
    PRECOND(start <= wave->samples, start = wave->samples);
    PRECOND(length <= wave->samples - start, length = wave->samples - start);

    /* If a length of zero was requested, return a valid pointer containing
     * no data (so the caller can distinguish between success and failure). */
    if (length == 0) {
        void *ptr = malloc(1);
        return ptr;
    }

    /* Allocate a buffer for the data. */
    float *buf = malloc(4 * wave->channels * length);
    if (!buf) {
        DMSG("Out of memory");
        return NULL;
    }

    /* Find the region corresponding to the start position. */
    unsigned int region = find_region(wave, start);

    /* Copy or read the data into the buffer. */
    uint64_t copied;
    uint64_t to_copy;  // Declare here so we can use it in the for() increment.
    for (copied = 0; copied < length; copied += to_copy, region++) {
        const uint64_t offset = start + copied - wave->regions[region].start;
        to_copy = ubound(length - copied,
                         wave->regions[region].length - offset);
        if (wave->regions[region].data) {
            /* Memory-backed data is already in floating-point format, so
             * we can just copy straight to the output buffer. */
            memcpy(&buf[wave->channels * copied],
                   &wave->regions[region].data[wave->channels * offset],
                   4 * wave->channels * to_copy);
        } else {
            /* Read the raw data from the file. */
            PRECOND(wave->file != NULL, free(buf); return NULL);
            const uint64_t file_offset =
                wave->regions[region].file_offset + offset;
            if (fseek(wave->file,
                      wave->data_offset + ((wave->bits/8) * wave->channels * file_offset),
                      SEEK_SET) != 0) {
                DMSG("fseek(): %s", strerror(errno));
                free(buf);
                return NULL;
            }
            uint64_t num_read = fread(&buf[wave->channels * copied],
                                      (wave->bits/8) * wave->channels, to_copy,
                                      wave->file);
            if (num_read < to_copy) {
                DMSG("Only read %llu of %llu samples from %llu",
                     (unsigned long long)num_read, (unsigned long long)to_copy,
                     (unsigned long long)file_offset);
                memset(&buf[wave->channels * (copied + num_read)], 0,
                       4 * wave->channels * (to_copy - num_read));
            }
            /* Convert the data to floating point. */
            float *out = &buf[wave->channels * copied];
            switch (wave->format) {
              case FMT_U8: {
                const uint8_t *in = (const uint8_t *)out;
                for (int64_t i = wave->channels * num_read - 1; i >= 0; i--) {
                    int32_t sample = (int8_t)(in[i] ^ 0x80);
                    out[i] = sample / 128.0f;
                }
                break;
              }
              case FMT_S16: {
                const uint8_t *in = (const uint8_t *)out;
                for (int64_t i = wave->channels * num_read - 1; i >= 0; i--) {
                    int32_t sample = (int8_t)in[i*2+1] << 8;
                    sample |= in[i*2+0] << 0;
                    out[i] = sample / 32768.0f;
                }
                break;
              }
              case FMT_S24: {
                const uint8_t *in = (const uint8_t *)out;
                for (int64_t i = wave->channels * num_read - 1; i >= 0; i--) {
                    int32_t sample = (int8_t)in[i*3+2] << 16;
                    sample |= in[i*3+1] << 8;
                    sample |= in[i*3+0] << 0;
                    out[i] = sample / 8388608.0f;
                }
                break;
              }
              case FMT_F32: {
                const uint8_t *in = (const uint8_t *)out;
                for (int64_t i = wave->channels * num_read - 1; i >= 0; i--) {
                    union {float f; uint32_t i;} sample;
                    sample.i =
                        in[i*4+3]<<24 | in[i*4+2]<<16 | in[i*4+1]<<8 | in[i*4];
                    out[i] = sample.f;
                }
                break;
              }
            }
        }  // if (wave->regions[region].data)
    }  // for (copied = 0; copied < length; copied += to_copy, region++)

    return buf;
}

/*-----------------------------------------------------------------------*/

/**
 * wave_insert_data:  Insert the given data into the given wave file.  The
 * data is assumed to have the same number of channels as the wave file.
 *
 * [Parameters]
 *         wave: Wave file into which to insert data
 *     position: Position at which to insert data, in samples
 *         data: Data to insert, as floating-point values
 *       length: Length of data to insert, in samples
 * [Return value]
 *     Nonzero on success, zero on error
 */
int wave_insert_data(WaveFile *wave, uint64_t position,
                     const float *data, uint64_t length)
{
    PRECOND(wave != NULL, return 0);
    PRECOND(position <= wave->samples, return 0);
    PRECOND(data != NULL, return 0);
    if (length == 0) {
        return 1;  // Not an error, but nothing to do.
    }

    /* If the insertion position is either in or immediately following a
     * memory-backed region, we can just insert (or append) the new data
     * directly into that region without modifying the region table. */
    unsigned int region = find_region(wave, position);
    if (region > 0 && wave->regions[region-1].data != NULL
                   && position == wave->regions[region-1].start
                                  + wave->regions[region-1].length) {
        region--;  // We'll insert (append) into the previous region
    }
    if (region < wave->num_regions && wave->regions[region].data != NULL) {
        const uint64_t new_length = wave->regions[region].length + length;
        const uint64_t offset = position - wave->regions[region].start;
        float *new_data = realloc(wave->regions[region].data,
                                  4 * wave->channels * new_length);
        if (!new_data) {
            DMSG("No memory for %llu samples", (unsigned long long)new_length);
            return 0;
        }
        if (offset < wave->regions[region].length) {
            memmove(&new_data[wave->channels * (offset + length)],
                    &new_data[wave->channels * offset],
                    4 * wave->channels * (wave->regions[region].length - offset));
        }
        memcpy(&new_data[wave->channels * offset], data,
               4 * wave->channels * length);
        wave->regions[region].length = new_length;
        wave->regions[region].data = new_data;
    } else {
        /* The insertion position is in the middle of file-backed data, so
         * we need to split at that position and insert a new memory-backed
         * region containing a copy of the passed-in data.  (We also reach
         * this case if the wave file is empty; in that case we treat it as
         * following a region of file-backed data and add the new data as
         * the first region.) */
        if (!split_region(wave, position)) {
            DMSG("Failed to split region at %llu",
                 (unsigned long long)position);
            return 0;
        }
        if (wave->num_regions >= MAX_REGIONS) {
            DMSG("No room to add a new region");
            return 0;
        }
        float *new_data = malloc(4 * wave->channels * length);
        if (!new_data) {
            DMSG("Out of memory");
            return 0;
        }
        memcpy(new_data, data, 4 * wave->channels * length);
        region = find_region(wave, position);
        if (region < wave->num_regions) {
            memmove(&wave->regions[region+1], &wave->regions[region],
                    sizeof(*wave->regions) * (wave->num_regions - region));
        }
        wave->num_regions++;
        wave->regions[region].length = length;
        wave->regions[region].data = new_data;
        wave->regions[region].file_offset = 0;  // Unused, but let's be clean

    }  // if (region < wave->num_regions && wave->regions[region].data != NULL)

    /* Update the region "start" fields before returning, since we inserted
     * data into the wave file. */
    update_region_starts(wave);

    /* Clear the full canvas data cache, since the wave data has been
     * modified. */
    free(wave->full_canvas_cache);
    wave->full_canvas_cache = NULL;
    wave->full_canvas_cache_width = 0;

    /* Adjust the current and saved position/selection so that they point
     * to the same samples they did before the insertion. */
    if (wave->position >= position) {
        wave->position += length;
    } else if (wave->position + wave->select_len >= position) {
        wave->select_len += length;
    }
    if (wave->saved_position >= position) {
        wave->saved_position += length;
    } else if (wave->saved_position + wave->saved_select_len > position) {
        wave->saved_select_len += length;
    }

    wave->samples += length;
    wave->modified = 1;
    return 1;
}

/*-----------------------------------------------------------------------*/

/**
 * wave_delete_data:  Delete the given section of the given wave file.
 *
 * [Parameters]
 *         wave: Wave file from which to delete data
 *     position: Position at which to delete data, in samples
 *       length: Length of data to delete, in samples
 * [Return value]
 *     Nonzero on success, zero on error
 */
int wave_delete_data(WaveFile *wave, uint64_t position, uint64_t length)
{
    PRECOND(wave != NULL, return 0);
    PRECOND(position <= wave->samples, return 0);
    PRECOND(length <= wave->samples - position,
            length = wave->samples - position);
    if (length == 0) {
        return 1;  // Not an error, but nothing to do.
    }

    /* Find the region containing the deletion starting point.  This will
     * always be a valid region, i.e. region will always be less than
     * wave->num_regions after this first find_region() call. */
    unsigned int region = find_region(wave, position);
    uint64_t offset = position - wave->regions[region].start;

    /* Truncate or delete regions whose endpoints are inside the section of
     * data to be deleted. */
    while (region < wave->num_regions
           && position + length >= wave->regions[region].start
                                   + wave->regions[region].length) {
        if (offset > 0) {
            wave->regions[region].length = offset;
            if (wave->regions[region].data) {
                float *new_data = realloc(
                    wave->regions[region].data,
                    4 * wave->channels * wave->regions[region].length);
                if (new_data) {
                    wave->regions[region].data = new_data;
                }
            }
            region++;
            offset = 0;
        } else {
            free(wave->regions[region].data);
            if (region+1 < wave->num_regions) {
                memmove(&wave->regions[region], &wave->regions[region+1],
                        sizeof(*wave->regions) * (wave->num_regions - (region+1)));
            }
            wave->num_regions--;
        }
    }

    /* If there is any remaining data to be deleted (i.e. not covering a
     * region endpoint), remove it. */
    if (region < wave->num_regions
     && position + length > wave->regions[region].start + offset
    ) {
        const uint64_t to_delete =
            (position + length) - (wave->regions[region].start + offset);
        if (wave->regions[region].data) {
            memmove(&wave->regions[region].data[wave->channels * offset],
                    &wave->regions[region].data[wave->channels * (offset + to_delete)],
                    4 * wave->channels * (wave->regions[region].length - (offset + to_delete)));
            wave->regions[region].length -= to_delete;
            float *new_data = realloc(
                wave->regions[region].data,
                4 * wave->channels * wave->regions[region].length);
            if (new_data) {
                wave->regions[region].data = new_data;
            }
        } else {
            if (offset > 0) {
                /* If we get to this point, the entire deletion region is
                 * contained within a single region (excluding the region's
                 * first and last samples); thus we can call split_region()
                 * without worrying that the region start fields are
                 * inaccurate, and we know that after split_region()
                 * succeeds, the region at "region" will in fact have been
                 * split into two parts. */
                if (!split_region(wave, position)) {
                    DMSG("Failed to split region at %llu for partial deletion",
                         (unsigned long long)position);
                    return 0;
                }
                wave->regions[region].length = offset;
                region++;
            }
            wave->regions[region].length -= to_delete;
            wave->regions[region].file_offset += to_delete;
        }
    }

    update_region_starts(wave);

    free(wave->full_canvas_cache);
    wave->full_canvas_cache = NULL;
    wave->full_canvas_cache_width = 0;

    /* Adjusting the position/selection after a deletion is more difficult
     * than for insertion, since there are more possibilities:
     *    1) Position is after the deleted region.
     *    2) Position is within the deleted region, and:
     *       A) Selection extends past the end of the deleted region.
     *       B) Selection is entirely contained within the deleted region.
     *    3) Position is before the deleted region, and:
     *       A) Selection extends past the end of the deleted region.
     *       B) Selection terminates within the deleted region.
     *       C) Selection terminates before the deleted region.
     * In every case but 3C, we have to adjust the position, selection
     * length, or both. */

    if (wave->position >= position + length) {
        wave->position -= length;
    } else if (wave->position >= position) {
        if (wave->position + wave->select_len >= position + length) {
            wave->select_len = (wave->position + wave->select_len)
                               - (position + length);
        } else {
            wave->select_len = 0;
        }
        wave->position = position;
    } else {
        if (wave->position + wave->select_len >= position + length) {
            wave->select_len -= length;
        } else if (wave->position + wave->select_len > position) {
            wave->select_len = position - wave->position;
        }
    }

    if (wave->saved_position > position + length) {
        wave->saved_position -= length;
    } else if (wave->saved_position > position) {
        if (wave->saved_position + wave->saved_select_len > position + length) {
            wave->saved_select_len =
                (wave->saved_position + wave->saved_select_len)
                - (position + length);
        } else {
            wave->saved_select_len = 0;
        }
        wave->saved_position = position;
    } else {
        if (wave->saved_position + wave->saved_select_len > position + length) {
            wave->saved_select_len -= length;
        } else if (wave->saved_position + wave->saved_select_len > position) {
            wave->saved_select_len = position - wave->saved_position;
        }
    }

    wave->samples -= length;
    wave->modified = 1;
    return 1;
}

/*-----------------------------------------------------------------------*/

/**
 * wave_take_data:  Replace the entire contents of the given wave file with
 * the given data.  The data is assumed to have the same number of channels
 * as the wave file.  The wave file is marked as modified by this routine.
 *
 * After calling this routine, the data buffer passed in is owned by the
 * wave file, and must not be modified or freed by the caller.
 *
 * This routine always succeeds if its preconditions are fulfilled.
 *
 * [Parameters]
 *       wave: Wave file into which to store data
 *       data: Data to store, as floating-point values
 *     length: Length of data to store, in samples
 * [Return value]
 *     None
 */
void wave_take_data(WaveFile *wave, float *data, uint64_t length)
{
    PRECOND(wave != NULL, return);
    PRECOND(data != NULL || length == 0, return);

    free(wave->path);
    if (wave->file) {
        fclose(wave->file);
    }
    free(wave->full_canvas_cache);
    wave->path              = NULL;
    wave->file              = NULL;
    wave->data_offset       = 0;
    wave->writable          = 1;
    wave->modified          = 1;
    wave->samples           = length;
    wave->full_canvas_cache = NULL;
    wave->full_canvas_cache_width = 0;
    wave->position          = 0;
    wave->select_len        = 0;
    wave->saved_position    = 0;
    wave->saved_select_len  = 0;
    wave->zoom              = 1;

    unsigned int region;
    for (region = 0; region < wave->num_regions; region++) {
        free(wave->regions[region].data);
        wave->regions[region].data = NULL;
    }
    if (length > 0) {
        wave->regions[0].start       = 0;
        wave->regions[0].length      = length;
        wave->regions[0].data        = data;
        wave->regions[0].file_offset = 0;
        wave->num_regions = 1;
    } else {
        free(data);
        wave->num_regions = 0;
    }
}

/*************************************************************************/

/**
 * wave_get_full_canvas:  Return data for drawing the entire waveform on a
 * canvas of the given width.  The data will be cached, so subsequent
 * requests for the same width will not require re-reading the data from
 * disk.
 *
 * The returned data is a sequence of unsigned 16-bit values, representing
 * the lowest and highest sample values for each channel in the range of
 * samples corresponding to each pixel:
 *    pixel0_ch0_low, pixel0_ch0_high, pixel0_ch1_low, ..., pixel1_ch0_low, ...
 * For each value, 0x0000 corresponds to the lowest possible sample value
 * (e.g. -0x8000 for signed 16-bit data), and 0xFFFF corresponds to the
 * highest possible sample value (e.g. +0x7FFF for signed 16-bit data).
 *
 * [Parameters]
 *      wave: Wave file for which to return data
 *     width: Width of canvas for which to return data
 * [Return value]
 *     Canvas drawing data, or NULL on error
 */
const uint16_t *wave_get_full_canvas(WaveFile *wave, unsigned int width)
{
    PRECOND(wave != NULL, return NULL);
    PRECOND(width != 0, return NULL);

    /* See if we have cached data.  If so, either return it (if the width is
     * unchanged) or discard it in preparation for regenerating the data. */
    if (wave->full_canvas_cache) {
        if (wave->full_canvas_cache_width == width) {
            return wave->full_canvas_cache;
        } else {
            free(wave->full_canvas_cache);
            wave->full_canvas_cache = NULL;
            wave->full_canvas_cache_width = 0;
        }
    }

    /* Allocate memory for the data we're about to generate. */
    uint16_t *canvas_data = malloc(2 * width * wave->channels * 2);
    if (!canvas_data) {
        DMSG("Out of memory (width = %u, channels = %u)",
             width, wave->channels);
        return NULL;
    }
    wave->full_canvas_cache_width = width;
    wave->full_canvas_cache = canvas_data;

    /* If the wave file is empty (zero samples), we can't process anything,
     * so just return zero data. */
    if (wave->samples == 0) {
        memset(canvas_data, 0, 2 * width * wave->channels * 2);
        return canvas_data;
    }

    /* Read through the file, finding the highest and lowest points on each
     * channel for each pixel's worth of data. */
    unsigned int x;
    for (x = 0; x < width; x++) {
        /* These are rounded up so the first pixel will always have at
         * least one sample's worth of data, and thus we won't need to
         * specially handle the case of no data for the first pixel. */
        const uint64_t position = (wave->samples * x + width-1) / width;
        const uint64_t length =
            ((wave->samples * (x+1) + width-1) / width) - position;
        uint16_t *data_ptr = &canvas_data[x * wave->channels * 2];
        if (length == 0) {
            /* There are no samples assigned to this pixel, so just copy
             * the last pixel's data. */
            memcpy(data_ptr, &canvas_data[(x-1) * wave->channels * 2],
                   2 * wave->channels * 2);
            continue;
        }
        find_high_low(wave, position, length, data_ptr);
    }

    return canvas_data;
}

/*************************************************************************/

/**
 * wave_get_channel_color:  Return a GdkColor for a channel in the given
 * waveform.  Colors are selected so that the same color represents the
 * same spatial position (left, center, LFE, and so on).
 *
 * [Parameters]
 *        wave: Wave file for which to return a color
 *     channel: Channel index (zero-based)
 * [Return value]
 *     Color for the given channel
 */
const GdkColor *wave_get_channel_color(const WaveFile *wave, int channel)
{
    static const GdkColor
        color_unknown      = {.pixel = 0xFF000000},  // Black
        color_front_left   = {.pixel = 0xFF0000FF},  // Blue
        color_front_right  = {.pixel = 0xFFFF0000},  // Red
        color_front_center = {.pixel = 0xFF00AA00},  // Green
        color_lfe          = {.pixel = 0xFFFFAA00},  // Orange
        color_rear_left    = {.pixel = 0xFF00AAFF},  // Cyan
        color_rear_right   = {.pixel = 0xFFFF00FF};  // Magenta

    PRECOND(wave != NULL, return &color_unknown);
    PRECOND(channel >= 0 && channel < wave->channels,
            return &color_unknown);

    switch (wave->channels) {
      case 1:  // Mono
        return &color_front_left;
      case 2:  // Stereo
        return channel==0 ? &color_front_left : &color_front_right;
      case 3:  // Stereo + center
        return channel==0 ? &color_front_left :
               channel==1 ? &color_front_center : &color_front_right;
      case 4:  // FL, FR, RL, RR
        return channel==0 ? &color_front_left :
               channel==1 ? &color_front_right :
               channel==2 ? &color_rear_left : &color_rear_right;
      case 5:  // FL, FR, C, RL, RR
        return channel==0 ? &color_front_left :
               channel==1 ? &color_front_right :
               channel==2 ? &color_front_center :
               channel==3 ? &color_rear_left : &color_rear_right;
      case 6:  // FL, FR, C, LFE, RL, RR
        return channel==0 ? &color_front_left :
               channel==1 ? &color_front_right :
               channel==2 ? &color_front_center :
               channel==3 ? &color_lfe :
               channel==4 ? &color_rear_left : &color_rear_right;
      default:
        return &color_unknown;
    }
};

/*************************************************************************/
/**************************** Local functions ****************************/
/*************************************************************************/

/**
 * find_region:  Return the index of the region corresponding to the given
 * sample index.
 *
 * [Parameters]
 *         wave: Wave file in which to look up the region
 *     position: Sample index to look up
 * [Return value]
 *     wave->regions[] index of region containing the sample at "position";
 *     returns wave->num_regions if position == wave->samples
 */
static unsigned int find_region(const WaveFile *wave, uint64_t position)
{
    PRECOND(wave != NULL, return 0);
    PRECOND(position <= wave->samples, return wave->num_regions);

    unsigned int region = 0;
    while (region < wave->num_regions
        && wave->regions[region].start + wave->regions[region].length <= position
    ) {
        region++;
    }
    return region;
}

/*-----------------------------------------------------------------------*/

/**
 * split_region:  Split the given region at the given sample index.  Does
 * nothing if "position" is the first sample in "region" or points to the
 * end of the wave file.
 *
 * [Parameters]
 *         wave: Wave file to operate on
 *     position: Sample index at which to split the region (counting from
 *                  the start of the wave file)
 * [Return value]
 *     Nonzero on success, zero on failure (region table full, out of
 *     memory, or invalid parameter)
 */
static int split_region(WaveFile *wave, uint64_t position)
{
    PRECOND(wave != NULL, return 0);
    PRECOND(position <= wave->samples, return 0);

    if (position == wave->samples) {
        /* Nothing to do. */
        return 1;
    }

    unsigned int region = find_region(wave, position);
    if (position == wave->regions[region].start) {
        /* Nothing to do. */
        return 1;
    }

    if (wave->num_regions >= MAX_REGIONS) {
        /* No room to add any more regions, so give up. */
        return 0;
    }

    /* Insert a new entry into the region list for the split data, and
     * update the old region's length. */
    if (region+1 < wave->num_regions) {
        memmove(&wave->regions[region+2], &wave->regions[region+1],
                sizeof(*wave->regions) * (wave->num_regions - (region+1)));
    }
    wave->num_regions++;
    wave->regions[region+1].start = position;
    wave->regions[region+1].length =
        wave->regions[region].start + wave->regions[region].length - position;
    const uint64_t offset = position - wave->regions[region].start;
    wave->regions[region].length = offset;

    /* Split the data buffer for the region, if appropriate. */
    if (wave->regions[region].data) {
        wave->regions[region+1].data =
            malloc(4 * wave->channels * wave->regions[region+1].length);
        if (!wave->regions[region+1].data) {
            DMSG("Out of memory");
            return 1;
        }
        memcpy(wave->regions[region+1].data,
               &wave->regions[region].data[wave->channels * offset],
               4 * wave->channels * wave->regions[region+1].length);
        /* Also try to shrink the old region's data; this should never fail,
         * but it won't cause any problems (other than memory pressure) if
         * it does, so we don't throw an error in any case. */
        float *shrunk_data = realloc(wave->regions[region].data,
                                     4 * wave->channels * offset);
        if (shrunk_data) {
            wave->regions[region].data = shrunk_data;
        }
    } else {
        wave->regions[region+1].data = NULL;
        wave->regions[region+1].file_offset =
            wave->regions[region].file_offset + offset;
    }

    /* No need to update the other region start pointers, since we haven't
     * actually inserted or deleted any audio data. */

    return 1;
}

/*-----------------------------------------------------------------------*/

/**
 * update_region_starts:  Update the "start" field of all regions in the
 * given wave file based on each region's length.  Used when the region
 * list has been modified.
 *
 * [Parameters]
 *     wave: Wave file to update
 * [Return value]
 *     None
 */
static void update_region_starts(WaveFile *wave)
{
    PRECOND(wave != NULL, return);

    uint64_t start = 0;
    unsigned int i;
    for (i = 0; i < wave->num_regions; i++) {
        wave->regions[i].start = start;
        start += wave->regions[i].length;
    }
}

/*************************************************************************/

/**
 * find_high_low:  Find the highest and lowest sample values for the given
 * sample range.  Helper function for wave_get_full_canvas().
 *
 * [Parameters]
 *         wave: Wave file to analyze
 *        start: Index of first sample to analyze
 *       length: Length of range to analyze, in samples
 *     data_ret: Pointer to array to receive the low and high values for
 *                  each channel, in order
 * [Return value]
 *     None
 */
static void find_high_low(const WaveFile *wave, uint64_t start,
                          uint64_t length, uint16_t *data_ret)
{
    PRECOND(wave != NULL, return);
    PRECOND(start <= wave->samples, return);
    PRECOND(length <= wave->samples - start, length = wave->samples - start);
    PRECOND(data_ret != NULL, return);

    /* Set the values so they'll always be updated when we process the
     * first pixel. */
    unsigned int ch;
    for (ch = 0; ch < wave->channels; ch++) {
        data_ret[ch*2+0] = 0xFFFF;
        data_ret[ch*2+1] = 0;
    }

    /* Find the region containing the first sample to process. */
    unsigned int region = find_region(wave, start);

    /* Analyze the sample data, region by region. */
    uint64_t i;
    uint64_t count;  // Declare here so we can use it in the for()
    for (i = 0; i < length; i += count, region++) {
        const uint64_t position = start + i;
        const uint64_t offset = position - wave->regions[region].start;
        count = ubound(length - i, wave->regions[region].length - offset);
        find_high_low_region(wave, region, offset, count, data_ret);
    }
}

/*----------------------------------*/

/**
 * find_high_low_region:  Find the highest and lowest sample values for the
 * given sample range in the given region.  Helper function for
 * find_high_low().
 *
 * [Parameters]
 *         wave: Wave file to analyze
 *       region: Index of region to analyze
 *       offset: Index of first sample in region to analyze (0 means the
 *                  first sample in _this region_, not in the file)
 *       length: Length of range to analyze, in samples
 *     data_ret: Pointer to array to receive the low and high values for
 *                  each channel, in order; must contain valid values on entry
 * [Return value]
 *     None
 */
static void find_high_low_region(const WaveFile *wave, unsigned int region,
                                 uint64_t offset, uint64_t length,
                                 uint16_t *data_ret)
{
    PRECOND(wave != NULL, return);
    PRECOND(region < wave->num_regions, return);
    PRECOND(wave->regions[region].data != NULL || wave->file != NULL, return);
    PRECOND(offset <= wave->regions[region].length, return);
    PRECOND(length <= wave->regions[region].length - offset,
            length = wave->regions[region].length - offset);
    PRECOND(data_ret != NULL, return);

    if (wave->regions[region].data) {

        /* Memory-backed regions are simple; since the sample format is
         * already known to be floating-point, we can just scan through
         * the data. */
        const float *src =
            &wave->regions[region].data[offset * wave->channels];
        uint64_t i;
        for (i = 0; i < length; i++) {
            unsigned int ch;
            for (ch = 0; ch < wave->channels; ch++, src++) {
                const uint16_t sample =
                    bound((int)roundf(*src * 32768), -32768, 32767) + 0x8000;
                data_ret[ch*2+0] = min(data_ret[ch*2+0], sample);
                data_ret[ch*2+1] = max(data_ret[ch*2+1], sample);
            }
        }

    } else {  // !wave->regions[region].data

        /* For file-backed regions, we read data in a chunk at a time to
         * avoid the overhead of numerous tiny read() calls. */

        uint8_t readbuf[49152];
        unsigned int readbuf_pos = sizeof(readbuf);

        const uint64_t file_offset =
            wave->data_offset + (wave->bits/8) * wave->channels
                                * (wave->regions[region].file_offset + offset);
        if (fseek(wave->file, file_offset, SEEK_SET) != 0) {
            DMSG("Failed to seek to %llu for sample %llu",
                 (unsigned long long)file_offset,
                 (unsigned long long)(wave->regions[region].start + offset));
            return;
        }

        uint64_t i;
        for (i = 0; i < length; i++) {
            unsigned int ch;
            for (ch = 0; ch < wave->channels; ch++) {
                if (readbuf_pos >= sizeof(readbuf)) {
                    const uint32_t num_read =
                        fread(readbuf, 1, sizeof(readbuf), wave->file);
                    if (num_read < sizeof(readbuf)) {
                        memset(&readbuf[num_read], 0,
                               sizeof(readbuf) - num_read);
                    }
                    readbuf_pos = 0;
                }
                uint16_t sample;
                switch (wave->format) {
                  case FMT_U8:
                    sample = readbuf[readbuf_pos]<<8 | readbuf[readbuf_pos];
                    readbuf_pos++;
                    break;
                  case FMT_S16:
                    sample = (readbuf[readbuf_pos] | readbuf[readbuf_pos+1]<<8)
                           + 0x8000;
                    readbuf_pos += 2;
                    break;
                  case FMT_S24:
                    sample = (readbuf[readbuf_pos+1] | readbuf[readbuf_pos+2]<<8)
                           + 0x8000;
                    readbuf_pos += 3;
                    break;
                  case FMT_F32: {
                    union {float f; uint32_t i;} value;
                    value.i = readbuf[readbuf_pos+0] <<  0
                            | readbuf[readbuf_pos+1] <<  8
                            | readbuf[readbuf_pos+2] << 16
                            | readbuf[readbuf_pos+3] << 24;
                    sample = bound((int)roundf(value.f*32768), -32768, 32767)
                           + 0x8000;
                    readbuf_pos += 4;
                    break;
                  }
                  default:
                    __builtin_unreachable();
                }
                data_ret[ch*2+0] = min(data_ret[ch*2+0], sample);
                data_ret[ch*2+1] = max(data_ret[ch*2+1], sample);
            }
        }

    }  // if (wave->regions[region].data)
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
